import logging
import os

from django.conf import settings
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.contrib.auth.models import Group as DjangoGroup

from .buildVOEvent import construct_voevent_file
from .models import Superevent, Log, Labelling, EMObservation, EMFootprint, \
    VOEvent, Signoff
from .shortcuts import is_superevent
from alerts.issuers.events import EventAlertIssuer, EventLogAlertIssuer
from alerts.issuers.superevents import SupereventAlertIssuer, \
    SupereventLogAlertIssuer, SupereventLabelAlertIssuer, \
    SupereventVOEventAlertIssuer, SupereventEMObservationAlertIssuer, \
    SupereventSignoffAlertIssuer, SupereventPermissionsAlertIssuer
from core.permissions import expose_log_to_lvem, expose_log_to_public, \
    hide_log_from_lvem, hide_log_from_public, assign_perms_to_obj, \
    remove_perms_from_obj
from core.vfile import create_versioned_file
from events.models import Event, EventLog, Tag, Label
from events.permission_utils import is_external
from events.shortcuts import is_event

# Set up logger
logger = logging.getLogger(__name__)

# NOTE: everything in here assumes that permissions have already been checked
# and handled properly

# Permissions which should be applied when exposing a superevent and removed
# when hiding it.  Keys are group names and values are permissions
# for that particulare group.
SUPEREVENT_PERMS = {
    settings.LVEM_OBSERVERS_GROUP: ['view', 'annotate'],
    settings.PUBLIC_GROUP: ['view'],
}


def create_superevent(submitter, t_start, t_0, t_end, preferred_event,
    events=[], labels=[], category='P', add_log_message=True,
    issue_alert=True):
    """
    Utility method for creating superevents.

    Arguments:
        submitter: User object (required)
        t_start, t_0, t_end: floats (required)
        preferred_event: Event object (required)
        events: list or QuerySet of Event objects
        labels: list of QuerySet of Label objects

    Usage:
        create_superevent(
            UserModel.objects.get(username='albert.einstein@ligo.org'),
            t_start=1, t_0=2, t_end=3, preferred_event=
            Event.objects.get(id='123400')
        )
    """

    # Create superevent
    s = Superevent.objects.create(submitter=submitter, t_start=t_start,
        t_0=t_0, t_end=t_end, preferred_event=preferred_event,
        category=category)

    # Create a log message to record initial superevent parameters
    creation_comment = ("Superevent created with t_start={t_start}, t_0={t_0},"
        " t_end={t_end}, preferred_event={preferred_event}").format(
        t_start=t_start, t_0=t_0, t_end=t_end, preferred_event=preferred_event)
    if events:
        creation_comment += ", events={events}".format(events=", ".join(
            [ev.graceid for ev in events]))
    # This is autogenerated, but people probably don't want to hide it
    # in the web interface
    superevent_creation_log = create_log(submitter, creation_comment, s,
        issue_alert=False, autogenerated=False)

    # Write a log message for preferred event
    pref_event_msg = "Set as preferred event for superevent: {superevent_id}" \
        .format(superevent_id=s.superevent_id)
    pref_event_log = create_log(submitter, pref_event_msg, preferred_event,
        issue_alert=False, autogenerated=False)

    # Add events to superevent
    # Superevent log message and alerts are taken care of elsewhere, but we
    # want to record logs for the individual events
    # NOTE: we don't have to worry about a repeat here for the preferred event
    # since the 'events' list comes directly from the serializer and hasn't
    # been updated to include the preferred event (like it would be if we
    # accessed s.events.all() directly)
    # Alerts aren't issued here since we want to do that *after* the superevent
    # creation alert is issued below.
    for event in events:
        add_event_to_superevent(s, event, submitter,
            add_superevent_log=False, add_event_log=True,
            issue_alert=False)

    # Create superevent data directory
    os.makedirs(s.datadir)

    # Issue all relevant alerts
    if issue_alert:
        # Send "new" alert about superevent creation
        SupereventAlertIssuer(s, alert_type='new').issue_alerts()

        # "Manually" issue alerts for preferred_event
        EventAlertIssuer(preferred_event, alert_type='selected_as_preferred') \
            .issue_alerts()

        # "Manually" issue alerts for events
        for event in events:
            EventAlertIssuer(event, alert_type='added_to_superevent') \
                .issue_alerts()

    # Add labels
    for label in labels:
        labelling, _ = add_label_to_superevent(s, label, submitter,
            add_log_message=True, issue_alert=issue_alert)

    return s


def update_superevent(superevent, updater, add_log_message=True,
    issue_alert=True, **kwargs):
    """
    kwargs which are used as superevent parameters:
        t_start, t_0, t_end, preferred_event
    """

    # Extract "updatable" superevent params from kwargs
    param_names = ['t_start', 't_0', 't_end', 'preferred_event']
    new_params = {k: v for k,v in kwargs.iteritems() if k in param_names}

    # Get old parameters
    old_params = {k: getattr(superevent, k) for k in new_params.keys()}

    # Update superevent object
    for k,v in new_params.iteritems():
        setattr(superevent, k, v)
    superevent.save()

    # Write a log message
    # 'if' statement to handle patch requests which don't change entire object
    superevent_alert_kwargs = {}
    if add_log_message:
        updates = ["{name}: {old} -> {new}".format(name=k, old=old_params[k],
            new=new_params[k]) for k in new_params.keys()
            if old_params[k] != new_params[k]]
        update_comment = "Updated superevent parameters: {0}".format(
            ", ".join(updates))
        update_log = create_log(updater, update_comment, superevent,
            issue_alert=False)

        # If preferred event changed, do a few things
        if new_params.has_key('preferred_event') and \
            (old_params['preferred_event'] != new_params['preferred_event']):
            # Write log for old preferred event
            old_msg = ("Removed as preferred event for superevent: "
                "{superevent_id}").format(superevent_id=
                superevent.superevent_id)
            old_log = create_log(updater, old_msg,
                    old_params['preferred_event'], issue_alert=False)

            # Write log for new preferred event
            new_msg = ("Set as preferred event for superevent: "
                "{superevent_id}").format(superevent_id=
                superevent.superevent_id)
            new_log = create_log(updater, new_msg,
                new_params['preferred_event'], issue_alert=False)

            # Issue alerts for both
            if issue_alert:
                # Old
                EventAlertIssuer(old_params['preferred_event'],
                    alert_type='removed_as_preferred').issue_alerts()
                # New
                EventAlertIssuer(new_params['preferred_event'],
                    alert_type='selected_as_preferred').issue_alerts()

        # Compile args for alert
        if 'preferred_event' in old_params:
            superevent_alert_kwargs['old_far'] = \
                old_params['preferred_event'].far
            superevent_alert_kwargs['old_nscand'] = \
                old_params['preferred_event'].is_ns_candidate()
        else:
            superevent_alert_kwargs['old_far'] = \
                superevent.preferred_event.far
            superevent_alert_kwargs['old_nscand'] = \
                superevent.preferred_event.is_ns_candidate()

    # Superevent alerts
    if issue_alert:
        SupereventAlertIssuer(superevent, alert_type='update').issue_alerts(
            **superevent_alert_kwargs)

    return superevent


def create_log(issuer, comment, event_or_superevent, filename="",
    data_file=None, tags=[], file_version=None, issue_alert=False,
    autogenerated=False):
    """
    Assume data have already been validated by a form or serializer.
    tags should be a list of Tag objects (??)
    data_file should be an in-memory file type
    """

    # Initial dictionary for log creation
    log_dict = {
        'issuer': issuer,
        'comment': comment,
        'filename': filename,
    }
    if file_version is not None:
        log_dict['file_version'] = file_version

    if is_superevent(event_or_superevent):
        log_dict['superevent'] = event_or_superevent
        log_attr = 'log_set'
        alerter_class = SupereventLogAlertIssuer
    elif is_event(event_or_superevent):
        log_dict['event'] = event_or_superevent
        log_attr = 'eventlog_set'
        alerter_class = EventLogAlertIssuer
    else:
        err_msg = "object is of type '{0}'; should be event or superevent" \
            .format(type(event_or_superevent))
        logger.error(err_msg)
        raise TypeError(err_msg)

    # Create log object
    log_set = getattr(event_or_superevent, log_attr)
    log = log_set.create(**log_dict)

    # Create versioned file
    if data_file:
        try:
            version = create_versioned_file(filename,
                event_or_superevent.datadir, data_file)
        except Exception as e:
            # If creating the file fails, we want to delete the log entry
            log.delete()
            raise e

        # Update file_version
        log.file_version = version
        log.save(update_fields=['file_version'])

    # Add tags to log messages
    for t in tags:
        add_tag_to_log(log, t, issuer, issue_alert=False)

    if issue_alert:
        alerter_class(log, alert_type='log').issue_alerts()

    # If user is external, add LV-EM tagname to this log message and expose it
    # TODO: should it be exposed to the public? Or to LV-EM only?  We will
    # stick with LV-EM only for now.
    if is_external(issuer) and not autogenerated:
        lvem_tag = Tag.objects.get(name=settings.EXTERNAL_ACCESS_TAGNAME)
        add_tag_to_log(log, lvem_tag, issuer, add_log_message=True)

    return log 


def get_log_parent(log):
    """Utility function to determine if this is an event or superevent log"""

    if isinstance(log, Log):
        return log.superevent
    elif isinstance(log, EventLog):
        return log.event
    else:
        err_msg = ("object is of type '{0}'; should be superevent log or "
            "event log").format(type(log))
        logger.error(err_msg)
        raise TypeError(err_msg)


def add_tag_to_log(log, tag, user, add_log_message=True, issue_alert=False):
    # Presently, we don't issue alerts for tag addition or for the logs
    # that are generated as a result.

    # Add tag to log
    log.tags.add(tag)

    # If this tag controls whether the log is exposed or not, we need to create
    # a corresponding GroupObjectPermission.  If we get to this point,
    # permissions should have already been checked.
    if (tag.name == settings.EXTERNAL_ACCESS_TAGNAME):
        expose_log_to_lvem(log)
    elif (tag.name == settings.PUBLIC_ACCESS_TAGNAME):
        expose_log_to_public(log)
        # Publicly exposed tags should also be exposed to LV-EM, if they
        # aren't already
        lvem_tag_applied = log.tags.filter(
            name=settings.EXTERNAL_ACCESS_TAGNAME).exists()
        if not lvem_tag_applied:
            lvem_tag = Tag.objects.get(name=settings.EXTERNAL_ACCESS_TAGNAME)
            add_tag_to_log(log, lvem_tag, user)

    # Create log message to record tag addition?
    log_for_tag_addition = None
    if add_log_message:
        comment = 'Tagged message {N}: {tag_name}.'.format(N=log.N,
            tag_name=tag.name)
        event_or_superevent = get_log_parent(log)
        log_for_tag_addition = create_log(user, comment, event_or_superevent,
            issue_alert=False, autogenerated=True)

    return log_for_tag_addition


def remove_tag_from_log(log, tag, user, add_log_message=True,
    issue_alert=False):
    # Presently, we don't issue alerts for tag addition or for the logs
    # that are generated as a result.

    # Remove tag from log
    log.tags.remove(tag)

    # If this tag controls whether the log is exposed or not, we need to create
    # a corresponding GroupObjectPermission.  If we get to this point,
    # permissions should have already been checked.
    if (tag.name == settings.EXTERNAL_ACCESS_TAGNAME):
        hide_log_from_lvem(log)
        # If the log is hidden from LV-EM, it should also be hidden from the
        # public
        public_tag_applied = log.tags.filter(
            name=settings.PUBLIC_ACCESS_TAGNAME).exists()
        if not public_tag_applied:
            public_tag = Tag.objects.get(name=settings.PUBLIC_ACCESS_TAGNAME)
            remove_tag_from_log(log, public_tag, user)
    elif (tag.name == settings.PUBLIC_ACCESS_TAGNAME):
        hide_log_from_public(log)

    # Create log message to record tag removal?
    log_for_tag_removal = None
    if add_log_message:
        comment = 'Removed tag {tag_name} from message {N}.'.format(
            N=log.N, tag_name=tag.name)
        event_or_superevent = get_log_parent(log)
        log_for_tag_removal = create_log(user, comment, event_or_superevent,
            issue_alert=False, autogenerated=True)

    return log_for_tag_removal


def add_event_to_superevent(superevent, event, user, add_event_log=True,
    add_superevent_log=True, issue_alert=True):

    # Check that the event is of the correct type to be added
    # to a superevent
    if not superevent.event_compatible(event):
        raise Superevent.EventCategoryMismatchError(
            _(('Event {graceid} is of type \'{e_category}\', and '
               'cannot be assigned to a superevent of type '
               '\'{s_category}\'').format(graceid=event.graceid,
                e_category=event.get_event_category(),
                s_category=superevent.get_category_display())))

    # Add event to superevent
    superevent.events.add(event)

    # Create superevent log message to record event addtion?
    if add_superevent_log:
        # Record event addition in superevent logs
        superevent_comment = 'Added event: {graceid}'.format(
            graceid=event.graceid)
        superevent_log_for_event_addition = create_log(user,
            superevent_comment, superevent, issue_alert=False,
            autogenerated=True)

    # Create event log message to record addition to superevent?
    if add_event_log:
        # Record addition to superevent in event logs
        event_comment = 'Added to superevent: {superevent_id}'.format(
            superevent_id=superevent.superevent_id)
        event_log_for_addition_to_superevent = create_log(user, event_comment,
            event, issue_alert=False, autogenerated=True)

    # Issue alerts
    if issue_alert:
        SupereventAlertIssuer(superevent, alert_type='event_added') \
            .issue_alerts()
        EventAlertIssuer(event, alert_type='added_to_superevent') \
            .issue_alerts()


def remove_event_from_superevent(superevent, event, user, add_event_log=True,
    add_superevent_log=True, issue_alert=True):
    """
    This function should be within a try-except block to catch exceptions and
    convert them to the appropriate response.
    """
    # Remove event from superevent
    superevent.events.remove(event)

    # Create superevent log message to record event removal?
    superevent_log_for_event_removal = None
    if add_superevent_log:
        # Record event removal in superevent logs
        superevent_comment = 'Removed event: {graceid}'.format(
            graceid=event.graceid)
        superevent_log_for_event_removal = create_log(user, superevent_comment,
            superevent, issue_alert=False, autogenerated=True)

    # Create event log message to record removal from superevent?
    if add_event_log:
        event_comment ='Removed from superevent: {superevent_id}'.format(
            superevent_id=superevent.superevent_id)
        event_log_for_removal_from_superevent = create_log(user, event_comment,
            event, issue_alert=False, autogenerated=True)

    # Issue alerts
    if issue_alert:
        SupereventAlertIssuer(superevent, alert_type='event_removed') \
            .issue_alerts()
        EventAlertIssuer(event, alert_type='removed_from_superevent') \
            .issue_alerts()


def add_label_to_superevent(superevent, label, user, add_log_message=True,
    issue_alert=True):

    # Create Labelling object - we use get_or_create due to the possibility
    # of IntegrityErrors. Sometimes two requests can get through the
    # before the first one is actually applied so get_or_create allows
    # us to safeguard against that.
    labelling, created = Labelling.objects.get_or_create(label=label,
        creator=user, superevent=superevent)

    # If the label is actually created (i.e. this isn't a secondary request
    # that squeaked through), then we optionally create a log message and
    # issue an alert.  If it's *not* created, then that stuff should have been
    # done already on a separate request and we don't need to repeat it here.
    # This is not ideal because users will get a response like they created
    # the label on that request but I don't think that is a significant
    # problem. This should be a rare occurrence and is basically here to
    # prevent server errors when users are doing dumb things.
    log_for_label_addition = None
    if created:
        if add_log_message:
            # Record label addition in superevent logs
            comment = 'Added label: {label_name}'.format(label_name=label.name)
            log_for_label_addition = create_log(user, comment, superevent,
                issue_alert=False, autogenerated=True)

        if issue_alert:
            SupereventLabelAlertIssuer(labelling, alert_type='label_added') \
                .issue_alerts()

    return labelling, log_for_label_addition


def remove_label_from_superevent(labelling, user, add_log_message=True,
    issue_alert=True):

    # Delete Labelling object
    labelling.delete()

    log_for_label_removal = None
    if add_log_message:
        # Record label addition in superevent logs
        comment = 'Removed label: {label_name}'.format(
            label_name=labelling.label.name)
        log_for_label_removal = create_log(user, comment, labelling.superevent,
            issue_alert=False, autogenerated=True)

    # labelling object still exists in memory, even though it has been
    # removed from the database and does not have an ID anymore
    if issue_alert:
        SupereventLabelAlertIssuer(labelling, alert_type='label_removed') \
            .issue_alerts()

    return log_for_label_removal


def get_or_create_tag(tag_name, display_name=None):

    tag, created = Tag.objects.get_or_create(name=tag_name)
    if created and display_name is not None:
        tag.displayName = display_name
        tag.save()

    return tag


def get_or_create_tags(tag_name_list, display_name_list=[]):

    # Check that lists are the same length if display_name_list is
    # provided
    if display_name_list and (len(display_name_list) != len(tag_name_list)):
        raise ValueError('If a list of display names is provided, it must '
            'have the same length as the list of tag names')

    tag_list = []
    for i, tag_name in enumerate(tag_name_list):
        display_name = None
        if display_name_list:
            display_name = display_name_list[i]
        tag = get_or_create_tag(tag_name, display_name)
        tag_list.append(tag)

    return tag_list


# TODO: move this somewhere else?
def get_superevent_by_date_id_or_404(superevent_id, queryset=None):

    try:
        filter_kwargs = Superevent.get_filter_kwargs_for_date_id_lookup(
            superevent_id)
    except Superevent.DateIdError as e:
        # The user passed an invalid date string (i.e., month=13
        # or something).  Probably should return 400 when this happens,
        # since it is technically a client error, but it's a lot simpler to
        # just raise a 404 here than to wrap every usage of this function in a
        # try-except block (or more than one, if it's in a CBV) and return a
        # 400. But raising a 404 is not technically wrong.
        raise Http404(e)

    if queryset is None:
        queryset = Superevent.objects.all()

    return get_object_or_404(queryset, **filter_kwargs)


def confirm_superevent_as_gw(superevent, user, add_log_message=True,
    issue_alert=True):

    # Save old ID temporarily
    old_id = superevent.superevent_id

    # Update superevent (mark as a GW, construct new ID, etc.)
    superevent.confirm_as_gw()

    # Create log message
    gw_log = None
    if add_log_message:
        message = ("Confirmed as a gravitational wave: ID changed from "
            "{old} -> {new}").format(old=old_id, new=superevent.superevent_id)
        gw_log = create_log(user, message, superevent, issue_alert=False,
            autogenerated=False)

    # Issue alert
    if issue_alert:
        SupereventAlertIssuer(superevent, alert_type='confirmed_as_gw') \
            .issue_alerts()

    return gw_log


def create_emobservation_for_superevent(superevent, submitter, ra_list,
    dec_list, ra_width_list, dec_width_list, start_time_list, duration_list,
    emgroup, comment="", add_log_message=True, issue_alert=True):

    # Create EMObservation
    emo = EMObservation.objects.create(submitter=submitter,
        superevent=superevent, group=emgroup, comment=comment)

    # Create EMFootprint objects
    for ra, dec, ra_width, dec_width, start_time, exposure_time in \
        zip(ra_list, dec_list, ra_width_list, dec_width_list, start_time_list,
        duration_list):

        emf = EMFootprint.objects.create(observation=emo, ra=ra, dec=dec,
            raWidth=ra_width, decWidth=dec_width, start_time=start_time,
            exposure_time=exposure_time)

    # Calculation covering region for observation
    emo.calculateCoveringRegion()
    emo.save()

    # Create a log message
    if add_log_message:
        message = "New EMBB observation record for {0}".format(emgroup.name)
        gw_log = create_log(submitter, message, superevent, issue_alert=False,
            autogenerated=False)

    # Issue alert
    if issue_alert:
        SupereventEMObservationAlertIssuer(emo, alert_type='emobservation') \
            .issue_alerts()

    return emo


def create_voevent_for_superevent(superevent, issuer, voevent_type,
    skymap_type=None, skymap_filename=None, internal=True, open_alert=False,
    hardware_inj=False, CoincComment=False, ProbHasNS=None,
    ProbHasRemnant=None, BNS=None, NSBH=None, BBH=None, Terrestrial=None,
    MassGap=None, add_log_message=True, issue_alert=True):

    # Instantiate VOEvent object
    voevent = VOEvent.objects.create(superevent=superevent, issuer=issuer,
        voevent_type=voevent_type, skymap_type=skymap_type,
        skymap_filename=skymap_filename, internal=internal,
        open_alert=open_alert, hardware_inj=hardware_inj,
        coinc_comment=CoincComment, prob_has_ns=ProbHasNS,
        prob_has_remnant=ProbHasRemnant, prob_bns=BNS, prob_nsbh=NSBH,
        prob_bbh=BBH, prob_terrestrial=Terrestrial, prob_mass_gap=MassGap)

    # Construct VOEvent file text
    voevent_text, ivorn = construct_voevent_file(superevent, voevent)

    # Save versioned VOEvent file
    voevent_display_type = dict(VOEvent.VOEVENT_TYPE_CHOICES) \
        [voevent.voevent_type].capitalize()
    voevent_filename = "{superevent}-{N}-{voevent_type}.xml".format(
        superevent=superevent.default_superevent_id, N=voevent.N,
        voevent_type=voevent_display_type)
    version = create_versioned_file(voevent_filename, superevent.datadir,
        voevent_text)

    # Update VOEvent object
    voevent.filename = voevent_filename
    voevent.file_version = version
    voevent.ivorn = ivorn
    voevent.save(update_fields=['filename', 'file_version', 'ivorn'])

    # Create a log entry to document the new VOEvent (tag it as em_follow)
    if add_log_message:
        comment = "New VOEvent"
        em_follow = Tag.objects.get(name='em_follow')
        voevent_log = create_log(issuer, comment, superevent,
            filename=voevent.filename, file_version=voevent.file_version,
            tags=[em_follow], issue_alert=False)

    # Issue an alert
    if issue_alert:
        SupereventVOEventAlertIssuer(voevent, alert_type='voevent') \
            .issue_alerts()

    return voevent


def create_signoff(superevent, user, signoff_type, signoff_instrument,
    signoff_status, signoff_comment, add_log_message=True, issue_alert=True):

    # Create signoff
    signoff = Signoff.objects.create(superevent=superevent, submitter=user,
        instrument=signoff_instrument, signoff_type=signoff_type,
        status=signoff_status, comment=signoff_comment)

    # Remove label which requested signoff
    labelling_to_remove = superevent.labelling_set.filter(label__name=
        signoff.get_req_label_name()).first()
    if labelling_to_remove is not None:
        remove_label_from_superevent(labelling_to_remove, user,
            add_log_message=False, issue_alert=False)

    # Add new label depending on signoff status
    label_to_add = Label.objects.get(name=signoff.get_status_label_name())
    add_label_to_superevent(superevent, label_to_add, user,
        add_log_message=False, issue_alert=issue_alert)

    # Create log message to document the signoff
    if add_log_message:

        # Compile a comment
        signoff_type_full = dict(Signoff.SIGNOFF_TYPE_CHOICES)[signoff_type]
        comment = "{signoff_type} signoff certified status as {status}".format(
            signoff_type=signoff_type_full.capitalize(), status=signoff_status)
        if signoff_instrument:
            comment += " for {inst}".format(inst=signoff_instrument)
        comment += ": '{comment}'".format(comment=signoff_comment)

        # Label-related comments
        label_removed_comment = ""
        if labelling_to_remove is not None:
            label_removed_comment = " label {l} removed and".format(
                l=labelling_to_remove.label.name)
        label_comment = ";{label_removed_comment} label {new_label} applied" \
            .format(label_removed_comment=label_removed_comment,
            new_label=label_to_add.name)
        # Add to full comment
        comment += label_comment

        # Add em_follow tag
        em_follow = Tag.objects.get(name='em_follow')

        # Create log
        signoff_log = create_log(user, comment, superevent, issue_alert=False,
            tags=[em_follow])

    # Issue alert
    if issue_alert:
        SupereventSignoffAlertIssuer(signoff, alert_type='signoff_created') \
            .issue_alerts()

    return signoff


def update_signoff(signoff, user, status, comment, add_log_message=True,
    issue_alert=True):
    # "Updatable" parameters for signoffs: status and comment
    new_data = {
        'status': status,
        'comment': comment,
    }
    updated_attributes = {k: v for k,v in new_data.items()
        if getattr(signoff, k, None) != v}
    old_data = {k: getattr(signoff, k) for k in updated_attributes}

    # Get superevent
    superevent = signoff.superevent

    # Update signoff values
    for k,v in updated_attributes.iteritems():
        setattr(signoff, k, v)
    signoff.save(update_fields=list(updated_attributes))

    # Manage labels
    if 'status' in updated_attributes:
        # If label for opposite status exists, remove it (i.e., the status has
        # changed in this update, so we need to update the labels)
        labelling_to_remove = superevent.labelling_set.filter(label__name=
            signoff.get_opposite_status_label_name()).first()
        if labelling_to_remove is not None:
            remove_label_from_superevent(labelling_to_remove, user,
                add_log_message=False, issue_alert=issue_alert)

        # If label for current status doesn't exist, add it
        label_to_add = Label.objects.get(name=signoff.get_status_label_name())
        if label_to_add not in superevent.labels.all():
            add_label_to_superevent(superevent, label_to_add, user,
                add_log_message=False, issue_alert=issue_alert)

    # Create log message to document the update
    if add_log_message:
        # Construct message
        signoff_type_full = dict(Signoff.SIGNOFF_TYPE_CHOICES) \
            [signoff.signoff_type]

        # Base comment string
        base_comment = "{signoff_type} signoff {for_inst}updated: {updates}"

        # Instrument-related string
        if signoff.instrument:
            comment_for_inst = "for {inst} ".format(inst=signoff.instrument)
        else:
            comment_for_inst = ""

        # Updates to specific parameters
        updates = []
        if 'status' in updated_attributes:
            # Careful with label removal in comment in case label doesn't
            # exist for some reason
            if labelling_to_remove is not None:
                r_label_comment = ', label {label} removed'.format(
                    label=labelling_to_remove.label.name)
            else:
                r_label_comment = ""
            status_comment = ("status {old_status} -> {new_status}"
                "{r_label_comment}, label {a_label} applied").format(
                old_status=signoff.opposite_status, new_status=signoff.status,
                r_label_comment=r_label_comment, a_label=label_to_add.name)
            updates.append(status_comment)
        if 'comment' in updated_attributes:
            comment_comment = "comment '{old}' -> '{new}'".format(
                old=old_data['comment'], new=updated_attributes['comment'])
            updates.append(comment_comment)

        # full comment
        full_comment_kwargs = {
            'signoff_type': signoff_type_full.capitalize(),
            'for_inst': comment_for_inst,
            'updates': "; ".join(updates)
        }
        full_comment = base_comment.format(**full_comment_kwargs)

        # Create log
        em_follow = Tag.objects.get(name='em_follow')
        signoff_log = create_log(user, full_comment, superevent,
            issue_alert=False, tags=[em_follow])

    # Issue alert
    if issue_alert:
        SupereventSignoffAlertIssuer(signoff, alert_type='signoff_updated') \
            .issue_alerts()

    return signoff


def delete_signoff(signoff, user, add_log_message=True,
    issue_alert=True):

    # Get superevent
    superevent = signoff.superevent

    # Delete signoff
    signoff.delete()

    # Remove current OK or NO label: we don't add a log message here since
    # we'll document this in the full log message about the signoff
    labelling_to_remove = superevent.labelling_set.filter(label__name=
        signoff.get_status_label_name()).first()
    remove_label_from_superevent(labelling_to_remove, user,
        add_log_message=False, issue_alert=True)

    # Reapply initial "REQ" or "OPS" label: we don't add a log message here
    # since we'll document this in the full log message about the signoff
    label_to_add = Label.objects.get(name=signoff.get_req_label_name())
    add_label_to_superevent(superevent, label_to_add, user,
        add_log_message=False, issue_alert=True)

    # Log message
    if add_log_message:
        # Construct log message
        signoff_type_full = dict(Signoff.SIGNOFF_TYPE_CHOICES) \
            [signoff.signoff_type]
        comment = "{signoff_type} signoff ".format(signoff_type=
            signoff_type_full.capitalize())
        if signoff.instrument:
            comment += "for {inst} ".format(inst=signoff.instrument)
        comment += "deleted: {r_label} removed and {a_label} reapplied".format(
            r_label=labelling_to_remove.label.name, a_label=label_to_add.name)
        em_follow = Tag.objects.get(name='em_follow')
        signoff_log = create_log(user, comment, superevent, tags=[em_follow],
            issue_alert=False)

    # Alert
    if issue_alert:
        SupereventSignoffAlertIssuer(signoff, alert_type='signoff_deleted') \
            .issue_alerts()


def expose_superevent(superevent, user, add_log_message=True,
    issue_alert=True):

    # Get groups
    lvem_group = DjangoGroup.objects.get(name=settings.LVEM_OBSERVERS_GROUP)
    public_group = DjangoGroup.objects.get(name=settings.PUBLIC_GROUP)

    # Assign permissions which will expose the superevent to LV-EM and the
    # public
    assign_perms_to_obj(SUPEREVENT_PERMS[lvem_group.name], lvem_group,
        superevent)
    assign_perms_to_obj(SUPEREVENT_PERMS[public_group.name], public_group,
        superevent)

    # Update superevent is_exposed attribute
    superevent.is_exposed = True
    superevent.save(update_fields=['is_exposed'])

    # Write log message
    if add_log_message:
        comment = 'Exposed to LV-EM observers and the general public.'
        create_log(user, comment, superevent, issue_alert=False)

    # Send alert
    if issue_alert:
        SupereventPermissionsAlertIssuer(superevent, alert_type='exposed') \
            .issue_alerts()


def hide_superevent(superevent, user, add_log_message=True,
    issue_alert=True):

    # Get groups
    lvem_group = DjangoGroup.objects.get(name=settings.LVEM_OBSERVERS_GROUP)
    public_group = DjangoGroup.objects.get(name=settings.PUBLIC_GROUP)

    # Assign permissions which will expose the superevent to LV-EM and the
    # public
    remove_perms_from_obj(SUPEREVENT_PERMS[lvem_group.name], lvem_group,
        superevent)
    remove_perms_from_obj(SUPEREVENT_PERMS[public_group.name], public_group,
        superevent)

    # Update superevent is_exposed attribute
    superevent.is_exposed = False
    superevent.save(update_fields=['is_exposed'])

    # Write log message
    if add_log_message:
        comment = 'Hidden from LV-EM observers and the general public.'
        create_log(user, comment, superevent, issue_alert=False)

    # Send alert
    if issue_alert:
        SupereventPermissionsAlertIssuer(superevent, alert_type='hidden') \
            .issue_alerts()
